from dataclasses import dataclass, asdict
from typing import Dict, List, Any


class MessageType:
    SCHEMA = 'SCHEMA'
    RECORD = 'RECORD'
    LINE = 'LINE'
    BOOKMARK = 'BOOKMARK'


@dataclass
class Message:
    type: str

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class SchemaMessage(Message):
    source: str
    stream: str
    schema: Dict[str, Any]
    key_properties: List[str]
    bookmark_properties: List[str]
    metadata: Dict[str, Any]

    def __init__(
        self,
        source: str,
        stream: str,
        schema: Dict[str, Any],
        key_properties: List[str],
        bookmark_properties: List[str],
        metadata: Dict[str, Any]
    ) -> None:
        self.type = MessageType.SCHEMA
        self.source = source
        self.stream = stream
        self.schema = schema
        self.key_properties = key_properties
        self.bookmark_properties = bookmark_properties
        self.metadata = metadata


@dataclass
class RecordMessage(Message):
    record: Dict[str, Any]

    def __init__(self, record: Dict[str, Any]):
        self.type = MessageType.RECORD
        self.record = record


@dataclass
class LineMessage(Message):
    line: str

    def __init__(self, line: str):
        self.type = MessageType.LINE
        self.line = line


@dataclass
class BookmarkMessage(Message):
    bookmark: Any

    def __init__(self, bookmark: Dict[str, Any]):
        self.type = MessageType.BOOKMARK
        self.bookmark = bookmark



import fastjsonschema
import logging
import sys
from typing import Iterator, Dict, List, Any, Optional, TextIO, Callable

from .infrastructure import default_loads, default_dumps
from .spec import (
    MessageType,
    SchemaMessage,
    RecordMessage,
    LineMessage,
    BookmarkMessage,
    Message
)


class Sink:
    def __init__(
            self,
            name: str,
            input_pipe: Optional[TextIO] = None,
            loads: Callable[[str], Any] = default_loads,
            dumps: Callable[[Any], str] = default_dumps
    ) -> None:
        self.name = name
        self.input_pipe = input_pipe or sys.stdin
        self.loads = loads
        self.dumps = dumps

        self.source: Optional[str] = None
        self.stream: Optional[str] = None
        self.schema: Dict[str, Any] = {}
        self.validate: Optional[Callable[[Any], None]] = None
        self.key_properties: List[str] = []
        self.bookmark_properties: List[str] = []
        self.metadata: Dict[str, Any] = {}
        self.bookmark: Dict[str, Any] = {}

    def get_bookmark(self, bookmark_property: str) -> Any:
        assert bookmark_property in self.bookmark_properties, f'{bookmark_property} not in bookmarks_properties'
        return self.bookmark.get(bookmark_property)

    def deserialize_message(self, line: str) -> Message:
        try:
            data: Dict[str, Any] = self.loads(line)
            msg_type = data.pop('type')
            if msg_type == MessageType.SCHEMA:
                return SchemaMessage(
                    source=data['source'],
                    stream=data['stream'],
                    schema=data['schema'],
                    key_properties=data['key_properties'],
                    bookmark_properties=data['bookmark_properties'],
                    metadata=data['metadata'])
            elif msg_type == MessageType.RECORD:
                return RecordMessage(
                    record=data['record'])
            elif msg_type == MessageType.LINE:
                return LineMessage(
                    line=data['line'])
            elif msg_type == MessageType.BOOKMARK:
                return BookmarkMessage(
                    bookmark=data['bookmark'])
            else:
                raise ValueError(f'Cannot read message: {data}')
        except Exception as exn:
            logging.error(f'failed to deserialize message: {line}\n{exn}')
            raise exn

    def read(self) -> Iterator[Message]:
        while True:
            line = self.input_pipe.readline()
            if line == "":
                return
            else:
                msg = self.deserialize_message(line)
                if isinstance(msg, SchemaMessage):
                    self.source = msg.source
                    self.stream = msg.stream
                    self.schema = msg.schema
                    self.validate = fastjsonschema.compile(msg.schema)
                    self.key_properties = msg.key_properties
                    self.bookmark_properties = msg.bookmark_properties
                    self.metadata = msg.metadata
                elif isinstance(msg, RecordMessage):
                    assert self.source is not None, 'source is not initialized'
                    assert self.stream is not None, 'stream is not initialized'
                    assert self.validate is not None, 'validate is not initialized'
                    try:
                        self.validate(msg.record)
                    except Exception as exn:
                        logging.error(f'validation failed for {msg.record}')
                        raise exn
                elif isinstance(msg, LineMessage):
                    assert self.source is not None, 'source is not initialized'
                    assert self.stream is not None, 'stream is not initialized'
                elif isinstance(msg, BookmarkMessage):
                    assert self.source is not None, 'source is not initialized'
                    assert self.stream is not None, 'stream is not initialized'
                    self.bookmark = msg.bookmark
                logging.debug(f'received message: {msg}')
                yield msg



import logging
import sys
from typing import Dict, List, Any, Optional, TextIO, Callable

from .infrastructure import default_loads, default_dumps
from .spec import (
    SchemaMessage,
    RecordMessage,
    LineMessage,
    BookmarkMessage,
    Message
)

DEFAULT_SCHEMA = {
    '$schema': 'http://json-schema.org/draft/2019-09/schema#',
    'type': 'object'
}


class Source:
    def __init__(
            self,
            name: str,
            bookmark: Optional[Dict[str, Any]] = None,
            stream: Optional[str] = None,
            schema: Optional[Dict[str, Any]] = None,
            key_properties: Optional[List[str]] = None,
            bookmark_properties: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            output_pipe: Optional[TextIO] = None,
            loads: Callable[[str], Any] = default_loads,
            dumps: Callable[[Any], str] = default_dumps
    ) -> None:
        self.name = name
        self.bookmark = bookmark or {}
        self.stream = stream
        self.schema = schema or DEFAULT_SCHEMA
        self.key_properties = key_properties or []
        self.bookmark_properties = bookmark_properties or []
        self.metadata = metadata or {}
        self.output_pipe = output_pipe or sys.stdout
        self.loads = loads
        self.dumps = dumps

    def _write(self, msg: Message) -> None:
        data = self.dumps(msg.to_dict()) + '\n'
        self.output_pipe.write(data)

    def get_bookmark(self, bookmark_property: str, default_value: Optional[Any] = None) -> Any:
        assert bookmark_property in self.bookmark_properties, f'{bookmark_property} not in bookmark_properties'
        return self.bookmark.get(bookmark_property, default_value)

    def update_bookmark(self, bookmark_property: str, bookmark_value: Any) -> None:
        assert bookmark_property in self.bookmark_properties, f'{bookmark_property} not in bookmark_properties'
        self.bookmark[bookmark_property] = bookmark_value

    def update_schema(
            self,
            stream: Optional[str] = None,
            schema: Optional[Dict[str, Any]] = None,
            key_properties: Optional[List[str]] = None,
            bookmark_properties: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None
    ) -> None:
        self.stream = stream or self.stream
        self.schema = schema or self.schema
        self.key_properties = key_properties or self.key_properties
        self.bookmark_properties = bookmark_properties or self.bookmark_properties
        self.metadata = metadata or self.metadata

    def write_schema(self) -> None:
        assert self.stream is not None, 'stream_name is undefined'
        msg = SchemaMessage(
            source=self.name,
            stream=self.stream,
            schema=self.schema,
            key_properties=self.key_properties,
            bookmark_properties=self.bookmark_properties,
            metadata=self.metadata)
        logging.info(f'writing schema {msg}')
        self._write(msg)

    def write_record(self, record: Dict[str, Any]) -> None:
        msg = RecordMessage(record=record)
        logging.debug(f'writing record {msg}')
        self._write(msg)

    def write_line(self, line: str) -> None:
        msg = LineMessage(line=line)
        logging.debug(f'writing line: {msg}')
        self._write(msg)

    def write_bookmark(self) -> None:
        msg = BookmarkMessage(bookmark=self.bookmark)
        logging.info(f'writing bookmark {msg}')
        self._write(msg)



import simplejson
from typing import Any


def default_loads(s: str) -> Any:
    return simplejson.loads(s, use_decimal=True)


def default_dumps(d: Any) -> str:
    return simplejson.dumps(d, use_decimal=True)



import redis
from typing import Callable, Any, Optional

from .bookmarker import Bookmarker
from ..infrastructure import default_dumps, default_loads


class RedisBookmarker(Bookmarker):
    def __init__(
        self,
        host: str = 'localhost',
        port: int = 6379,
        password: Optional[str] = None,
        database: int = 0,
        loads: Callable[[str], Any] = default_loads,
        dumps: Callable[[Any], str] = default_dumps
    ) -> None:
        self.redis = redis.Redis(
            host=host,
            port=port,
            db=database,
            password=password)
        self.loads = loads
        self.dumps = dumps

    def get_bookmark(self, source: str, stream: str, sink: str) -> Any:
        key = f'{source}:{stream}:{sink}'
        if (data := self.redis.get(key)) is not None:
            return self.loads(data)

    def set_bookmark(self, source: str, stream: str, sink: str, value: Any) -> None:
        key = f'{source}:{stream}:{sink}'
        data = self.dumps(value)
        self.redis.set(key, data)



from .spec import (
    Message,
    SchemaMessage,
    RecordMessage,
    LineMessage,
    BookmarkMessage
)
from .bookmarker import Bookmarker, RedisBookmarker
from .source import Source
from .sink import Sink
__all__ = [
    'Message',
    'SchemaMessage',
    'RecordMessage',
    'LineMessage',
    'BookmarkMessage',
    'Source',
    'Sink',
    'Bookmarker',
    'RedisBookmarker'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['etlaas_stream', 'etlaas_stream.bookmarker']

package_data = \
{'': ['*']}

install_requires = \
['fastjsonschema>=2.14.4,<3.0.0',
 'redis>=3.5.3,<4.0.0',
 'simplejson>=3.17.0,<4.0.0']

setup_kwargs = {
    'name': 'etlaas-stream',
    'version': '20.7.4',
    'description': 'Convenience classes for working with data streams.',
    'long_description': None,
    'author': 'Andrew Meier',
    'author_email': 'ameier38@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any


class Bookmarker:
    def get_bookmark(self, source: str, stream: str, sink: str) -> Any:
        """
        Get the source's stream bookmark for the target.
        :param source: Name of the source.
        :param stream: Name of the stream.
        :param sink: Name of the sink.
        :return: bookmark
        """
        raise NotImplementedError()

    def set_bookmark(self, source: str, stream: str, sink: str, value: Any) -> None:
        """
        Set the source's stream bookmark for the target.
        :param source: Name of the source.
        :param stream: Name of the stream.
        :param sink: Name of the sink.
        :param value: Bookmark value.
        :return: None
        """
        raise NotImplementedError()


